@@ -7,13 +7,15 @@
 /*	#include <dns.h>
 /*
 /*	DNS_RR	*dns_rr_create(qname, rname, type, class, ttl, preference,
-/*				data, data_len)
+/*				weight, port, data, data_len)
 /*	const char *qname;
 /*	const char *rname;
 /*	unsigned short type;
 /*	unsigned short class;
 /*	unsigned int ttl;
 /*	unsigned preference;
+/*	unsigned weight;
+/*	unsigned port;
 /*	const char *data;
 /*	size_t data_len;
 /*
@@ -27,6 +29,10 @@
 /*	DNS_RR	*list;
 /*	DNS_RR	*record;
 /*
+/*	DNS_RR	*dns_rr_append_discard(list, record)
+/*	DNS_RR	*list;
+/*	DNS_RR	*record;
+/*
 /*	DNS_RR	*dns_rr_sort(list, compar)
 /*	DNS_RR	*list
 /*	int	(*compar)(DNS_RR *, DNS_RR *);
@@ -52,6 +58,29 @@
 /*
 /*	DNS_RR	*dns_srv_rr_sort(list)
 /*	DNS_RR	*list;
+/*
+/*	int	var_dns_rr_list_limit;
+/* AUXILIARY FUNCTIONS
+/*	DNS_RR	*dns_rr_create_nopref(qname, rname, type, class, ttl,
+/*					data, data_len)
+/*	const char *qname;
+/*	const char *rname;
+/*	unsigned short type;
+/*	unsigned short class;
+/*	unsigned int ttl;
+/*	const char *data;
+/*	size_t data_len;
+/*
+/*	DNS_RR	*dns_rr_create_noport(qname, rname, type, class, ttl,
+/*					preference, data, data_len)
+/*	const char *qname;
+/*	const char *rname;
+/*	unsigned short type;
+/*	unsigned short class;
+/*	unsigned int ttl;
+/*	unsigned preference;
+/*	const char *data;
+/*	size_t data_len;
 /* DESCRIPTION
 /*	The routines in this module maintain memory for DNS resource record
 /*	information, and maintain lists of DNS resource records.
@@ -59,10 +88,14 @@
 /*	dns_rr_create() creates and initializes one resource record.
 /*	The \fIqname\fR field specifies the query name.
 /*	The \fIrname\fR field specifies the reply name.
-/*	\fIpreference\fR is used for MX records; \fIdata\fR is a null
+/*	\fIpreference\fR is used for MX and SRV records; \fIweight\fR
+/*	and \fIport\fR are used for SRV records; \fIdata\fR is a null
 /*	pointer or specifies optional resource-specific data;
 /*	\fIdata_len\fR is the amount of resource-specific data.
 /*
+/*	dns_rr_create_nopref() and dns_rr_create_noport() are convenience
+/*	wrappers around dns_rr_create() that take fewer arguments.
+/*
 /*	dns_rr_free() releases the resource used by of zero or more
 /*	resource records.
 /*
@@ -72,6 +105,16 @@
 /*	record(s).
 /*	A null input list is explicitly allowed.
 /*
+/*	dns_rr_append_discard() appends of discards a resource
+/*	record list. When the existing list already contains
+/*	var_dns_rr_list_limit elements (default: 100),
+/*	dns_rr_append_discard() logs a warning, sets the existing
+/*	list length to the sentinel value DNS_RR_DISCARDED and
+/*	discards the new resource record(s) instead of appending
+/*	them. Once a list length is set to DNS_RR_DISCARDED, the
+/*	caller is expected to stop trying to append records to that
+/*	list.
+/*
 /*	dns_rr_sort() sorts a list of resource records into ascending
 /*	order according to a user-specified criterion. The result is the
 /*	sorted list.
@@ -102,7 +145,7 @@
 /*	111 8th Avenue
 /*	New York, NY 10011, USA
 /*
-/*	SRV Support by 
+/*	SRV Support by
 /*	Tomas Korbar
 /*	Red Hat, Inc.
 /*--*/
@@ -123,6 +166,16 @@
 
 #include "dns.h"
 
+ /*
+  * A generous safety limit for the number of DNS resource records that the
+  * Postfix DNS client library will admit into a list. The default value 100
+  * is 20x the default limit on the number address records that the Postfix
+  * SMTP client is willing to consider.
+  * 
+  * Mutable, to make code testable.
+  */
+int     var_dns_rr_list_limit = 100;
+
 /* dns_rr_create - fill in resource record structure */
 
 DNS_RR *dns_rr_create(const char *qname, const char *rname,
@@ -133,7 +186,10 @@ DNS_RR *dns_rr_create(const char *qname, const char *rname,
 {
     DNS_RR *rr;
 
-    rr = (DNS_RR *) mymalloc(sizeof(*rr) + data_len - 1);
+    /*
+     * Note: if this function is changed, update dns_rr_copy().
+     */
+    rr = (DNS_RR *) mymalloc(sizeof(*rr));
     rr->qname = mystrdup(qname);
     rr->rname = mystrdup(rname);
     rr->type = type;
@@ -143,10 +199,15 @@ DNS_RR *dns_rr_create(const char *qname, const char *rname,
     rr->pref = pref;
     rr->weight = weight;
     rr->port = port;
-    if (data && data_len > 0)
+    if (data_len != 0) {
+	rr->data = mymalloc(data_len);
 	memcpy(rr->data, data, data_len);
+    } else {
+	rr->data = 0;
+    }
     rr->data_len = data_len;
     rr->next = 0;
+    rr->len = 1;
     return (rr);
 }
 
@@ -159,6 +220,8 @@ void    dns_rr_free(DNS_RR *rr)
 	    dns_rr_free(rr->next);
 	myfree(rr->qname);
 	myfree(rr->rname);
+	if (rr->data)
+	    myfree(rr->data);
 	myfree((void *) rr);
     }
 }
@@ -167,16 +230,17 @@ void    dns_rr_free(DNS_RR *rr)
 
 DNS_RR *dns_rr_copy(DNS_RR *src)
 {
-    ssize_t len = sizeof(*src) + src->data_len - 1;
     DNS_RR *dst;
 
     /*
-     * Combine struct assignment and data copy in one block copy operation.
+     * Note: struct copy, because dns_rr_create() would not copy all fields.
      */
-    dst = (DNS_RR *) mymalloc(len);
-    memcpy((void *) dst, (void *) src, len);
+    dst = (DNS_RR *) mymalloc(sizeof(*dst));
+    *dst = *src;
     dst->qname = mystrdup(src->qname);
     dst->rname = mystrdup(src->rname);
+    if (dst->data)
+	dst->data = mymemdup(src->data, src->data_len);
     dst->next = 0;
     return (dst);
 }
@@ -185,10 +249,43 @@ DNS_RR *dns_rr_copy(DNS_RR *src)
 
 DNS_RR *dns_rr_append(DNS_RR *list, DNS_RR *rr)
 {
+    if (rr == 0)
+	return (list);
     if (list == 0) {
 	list = rr;
     } else {
 	list->next = dns_rr_append(list->next, rr);
+	/* Non-sentinel lengths are o(100) an are always safe to add. */
+	if (list->len == DNS_RR_DISCARDED || rr->len == DNS_RR_DISCARDED)
+	    list->len = DNS_RR_DISCARDED;
+	else
+	    list->len += rr->len;
+    }
+    return (list);
+}
+
+/* dns_rr_append_discard - append resource record to list, or discard */
+
+DNS_RR *dns_rr_append_discard(DNS_RR *list, DNS_RR *rr)
+{
+    if (rr == 0)
+	return (list);
+    if (list == 0) {
+	list = rr;
+    } else if (list->len < var_dns_rr_list_limit) {
+	list->next = dns_rr_append(list->next, rr);
+	/* Non-sentinel lengths are o(100) an are always safe to add. */
+	if (rr->len == DNS_RR_DISCARDED)
+	    list->len = DNS_RR_DISCARDED;
+	else
+	    list->len += rr->len;
+    } else {
+	if (list->len != DNS_RR_DISCARDED) {
+	    msg_warn("dropping excess records for qname=%s qtype=%s",
+		     rr->qname, dns_strtype(rr->type));
+	    list->len = DNS_RR_DISCARDED;
+	}
+	dns_rr_free(rr);
     }
     return (list);
 }
@@ -265,6 +362,12 @@ DNS_RR *dns_rr_sort(DNS_RR *list, int (*compar) (DNS_RR *, DNS_RR *))
     int     len;
     int     i;
 
+    /*
+     * Avoid mymalloc() panic.
+     */
+    if (list == 0)
+	return (list);
+
     /*
      * Save state and initialize.
      */
@@ -311,6 +414,12 @@ DNS_RR *dns_rr_shuffle(DNS_RR *list)
     int     i;
     int     r;
 
+    /*
+     * Avoid mymalloc() panic.
+     */
+    if (list == 0)
+	return (list);
+
     /*
      * Build linear array with pointers to each list element.
      */
@@ -364,55 +473,50 @@ DNS_RR *dns_rr_remove(DNS_RR *list, DNS_RR *record)
     return (list);
 }
 
+/* weight_order - sort equal-priority records by weight */
+
 static void weight_order(DNS_RR **array, int count)
 {
-    /* Compute sum of weights. */
-    int     weight_sum = 0;
+    int     unordered_weights;
     int     i;
 
-    for (i = 0; i < count; i++)
-	weight_sum += array[i]->weight;
-
-    /* If weights are not supplied then we do not have to order records. */
-    if (weight_sum == 0)
+    /*
+     * Compute the sum of record weights. If weights are not supplied then
+     * this function would be a noop. In fact this would be a noop when all
+     * weights have the same value, whether that weight is zero or not. There
+     * is no need to give special treatment to zero weights.
+     */
+    for (unordered_weights = 0, i = 0; i < count; i++)
+	unordered_weights += array[i]->weight;
+    if (unordered_weights == 0)
 	return;
 
-    /* First move records with weight 0 to the beginning. */
-    int     swap_place = 0;
-    DNS_RR *temp;
-
-    for (i = 0; i < count; i++) {
-	if (array[i]->weight == 0) {
-	    temp = array[swap_place];
-	    array[swap_place] = array[i];
-	    array[i] = temp;
-	    swap_place++;
-	}
-    }
-
-    int     random;
-
-    unsigned int running_sums[count];
-
+    /*
+     * The record ordering code below differs from RFC 2782 when the input
+     * contains a mix of zero and non-zero weights: the code below does not
+     * give special treatment to zero weights. Instead, it treats a zero
+     * weight just like any other small weight. Fewer special cases make for
+     * code that is simpler and more robust.
+     */
     for (i = 0; i < count - 1; i++) {
-	running_sums[i] = array[i]->weight;
-	/* Calculate running sums of records. */
-	int     x;
-
-	for (x = i + 1; x < count; x++)
-	    running_sums[x] = array[x]->weight + running_sums[x - 1];
-
-	random = myrand() % (weight_sum + 1);
+	int     running_sum;
+	int     threshold;
+	int     k;
+	DNS_RR *temp;
 
 	/*
-	 * Find first record that has running sum greater or equal to the
-	 * random number
+	 * Choose a random threshold [0..unordered_weights] inclusive.
 	 */
-	int     k;
+	threshold = myrand() % (unordered_weights + 1);
 
-	for (k = i; k < count; k++) {
-	    if (running_sums[k] >= random) {
-		weight_sum -= array[k]->weight;
+	/*
+	 * Move the first record with running_sum >= threshold to the ordered
+	 * list, and update unordered_weights.
+	 */
+	for (running_sum = 0, k = i; k < count; k++) {
+	    running_sum += array[k]->weight;
+	    if (running_sum >= threshold) {
+		unordered_weights -= array[k]->weight;
 		temp = array[i];
 		array[i] = array[k];
 		array[k] = temp;
@@ -432,6 +536,15 @@ DNS_RR *dns_srv_rr_sort(DNS_RR *list)
     int     len;
     int     i;
     int     r;
+    int     cur_pref;
+    int     left_bound;			/* inclusive */
+    int     right_bound;		/* non-inclusive */
+
+    /*
+     * Avoid mymalloc() panic, or rr_array[0] fence-post error.
+     */
+    if (list == 0)
+	return (list);
 
     /*
      * Save state and initialize.
@@ -463,29 +576,23 @@ DNS_RR *dns_srv_rr_sort(DNS_RR *list)
     /* First order the records by preference. */
     qsort((void *) rr_array, len, sizeof(*rr_array), dns_rr_sort_callback);
 
-    int     cur_pref = rr_array[0]->pref;
-    int     left_bound = 0;
-    int     right_bound = 0;
-
     /*
-     * Walk through records and sort every partition with the same
-     * preference, according to their weight.
+     * Walk through records and sort the records in every same-preference
+     * partition according to their weight. Note that left_bound is
+     * inclusive, and that right-bound is non-inclusive.
      */
-    int     k;
+    left_bound = 0;
+    cur_pref = rr_array[left_bound]->pref;	/* assumes len > 0 */
 
-    for (k = 1; k < len; k++) {
-	if (rr_array[k]->pref != cur_pref) {
-	    if (left_bound != right_bound) {
-		weight_order(&rr_array[left_bound], right_bound - left_bound + 1);
-	    }
-	    cur_pref = rr_array[k]->pref;
-	    left_bound = k;
+    for (right_bound = 1; /* see below */ ; right_bound++) {
+	if (right_bound == len || rr_array[right_bound]->pref != cur_pref) {
+	    if (right_bound - left_bound > 1)
+		weight_order(rr_array + left_bound, right_bound - left_bound);
+	    if (right_bound == len)
+		break;
+	    left_bound = right_bound;
+	    cur_pref = rr_array[left_bound]->pref;
 	}
-	right_bound = k;
-    }
-    /* We need to check whether there is not one more partition to sort. */
-    if (left_bound != right_bound) {
-	weight_order(&rr_array[left_bound], right_bound - left_bound + 1);
     }
 
     /*
